@@ -2,7 +2,17 @@
 pyndl.wh
 --------
 
-*pyndl.wh* provides functions in order to train WH models
+*pyndl.wh* provides functions in order to train Widrow-Hoff (WH) models. In contrast
+to the Rescorla-Wagner (RW) models, the WH models can not only have binary
+cues and outcomes, but can encode gradual intensities in the cues and outcomes.
+This is done by associating a vector of continues values (real numbers) to each
+cue and outcome. The size of the vector has to be the same for all cues and for
+all outcomes, but can differ between cues and outcomes.
+
+It is possible to calculate weights for continuous cues or continues outcomes,
+while keeping the outcomes respectively cues binary. Finally, it is possible to
+have both sides, cues and outcomes, to be continues and calculate the
+Widrow-Hoff learning rule between them.
 
 """
 from collections import defaultdict, OrderedDict
@@ -45,11 +55,20 @@ def wh(events, eta, *, cue_vectors=None, outcome_vectors=None,
         verbose=False, temporary_directory=None,
         events_per_temporary_file=10000000):
     """
-    Calculate the weights for all events using the Widrow-Hoff learning rule
-    and training as outcomes on sematic vectors in semantics.
+    Calculate the weights for all events using the Widrow-Hoff learning rule in
+    three different flavors.
 
-    This is a parallel python implementation using numpy, multithreading and
-    the binary format defined in preprocess.py.
+    In the first flavor, cues and outcomes both are vectors and the names in
+    the eventfiles refer to these vectors. The vectors for all cues and
+    outcomes are given as an xarray.DataArray with the arguments `cue_vectors`
+    and `outcome_vectors'.
+
+    In the second and third flavor, only the cues or only the outcomes are
+    treated as vectors and the ones not being treated as vectors are still
+    considered being present or not being present in a binary way.
+
+    This is a parallel python implementation using cython, numpy,
+    multithreading and the binary format defined in preprocess.py.
 
     Parameters
     ----------
@@ -64,23 +83,24 @@ def wh(events, eta, *, cue_vectors=None, outcome_vectors=None,
     method : {'openmp', 'threading', 'numpy'}
         'numpy' works only for real to real Widrow-Hoff.
     weights : None or xarray.DataArray
-        the xarray.DataArray needs to have the dimensions 'cues' and 'outcomes'
+        the xarray.DataArray needs to have the named dimensions 'cues' or
+        'cue_vector_dimensions' and 'outcomes' or 'outcome_vector_dimensions'
     n_jobs : int
         an integer giving the number of threads in which the job should be
         executed
     n_outcomes_per_job : int
         an integer giving the number of outcomes that are processed in one job
     remove_duplicates : {None, True, False}
-        if None though a ValueError when the same cue is present multiple times
+        if None raise a ValueError when the same cue is present multiple times
         in the same event; True make cues and outcomes unique per event; False
         keep multiple instances of the same cue or outcome (this is usually not
         preferred!)
     verbose : bool
-        print some output if True.
+        print some output if True
     temporary_directory : str
         path to directory to use for storing temporary files created;
         if none is provided, the operating system's default will
-        be used (/tmp on unix)
+        be used like '/tmp' on unix
     events_per_temporary_file: int
         Number of events in each temporary binary file. Has to be larger than 1
 
@@ -91,12 +111,11 @@ def wh(events, eta, *, cue_vectors=None, outcome_vectors=None,
         run (real to real, binary to real, real to binary or binary to binary).
         The dimension names reflect this in the weights. They are a combination
         of 'outcomes' x 'outcome_vector_dimensions' and 'cues' x
-        'cue_vector_dimensions'
-        with dimensions 'outcome_vector dimensions' and
-        'cue_vector_dimensions'. You can lookup the weights
-        between a vector dimension and a cue with
-        ``weights.loc[{'outcome_vector_dimensions': outcome_vector_dimension,
-        'cue_vector_dimensions': cue_vector_dimension}]`` or
+        'cue_vector_dimensions' with dimensions 'outcome_vector dimensions' and
+        'cue_vector_dimensions'. You can lookup the weights between a vector
+        dimension and a cue with ``weights.loc[{'outcome_vector_dimensions':
+        outcome_vector_dimension, 'cue_vector_dimensions':
+        cue_vector_dimension}]`` or
         ``weights.loc[vector_dimension].loc[cue_vector_dimension]``.
 
     """
@@ -150,6 +169,16 @@ def dict_wh(events, eta, cue_vectors, outcome_vectors, *,
     `dict_ndl` cannot be used to continue learning. At the moment there is no
     proper way to automatically store the meta data into the default dict.
 
+    Furthermore, this implementation only supports the 'real to real' case
+    where cue vectors are learned on outcome vectors. For the 'binary to real'
+    or 'real to binary' cases the `wh.wh` function needs to be used which uses
+    a fast cython implementation.
+
+    The main purpose of this function is to have a reference implementation
+    which is used to validate the faster cython version against. Additionally, this
+    function can be a good starting point to develop different flavors of the
+    Widrow-Hoff learning rule.
+
     Parameters
     ----------
     events : generator or str
@@ -185,9 +214,12 @@ def dict_wh(events, eta, cue_vectors, outcome_vectors, *,
     or
 
     weights : xarray.DataArray
-        with dimensions 'outcomes' and 'cues'. You can lookup the weights
-        between a cue and an outcome with ``weights.loc[{'outcomes': outcome,
-        'cues': cue}]`` or ``weights.loc[outcome].loc[cue]``.
+        with dimensions 'outcome_vector_dimensions' and
+        'cue_vector_dimensions'. You can lookup the weights
+        between a cue dimension and an outcome dimension with
+        ``weights.loc[{'outcome_vector_dimensions': outcome_vector_dimension,
+        'cue_vector_dimensions': cue_vector_dimension}]`` or
+        ``weights.loc[outcome_vector_dimension].loc[cue_vector_dimension]``.
 
     """
 
@@ -305,10 +337,10 @@ def _wh_binary_to_real(events, eta, outcome_vectors, *,
                        events_per_temporary_file=10000000):
     """
     Calculate the weights for all events using the Widrow-Hoff learning rule
-    and training as outcomes on sematic vectors in semantics.
+    from binary cues onto continuous outcome vectors.
 
-    This is a parallel python implementation using numpy, multithreading and
-    the binary format defined in preprocess.py.
+    This is a parallel python implementation using cython, numpy,
+    multithreading and the binary format defined in preprocess.py.
 
     Parameters
     ----------
@@ -321,7 +353,8 @@ def _wh_binary_to_real(events, eta, outcome_vectors, *,
 
     method : {'openmp', 'threading'}
     weights : None or xarray.DataArray
-        the xarray.DataArray needs to have the dimensions 'cues' and 'outcomes'
+        the xarray.DataArray needs to have the dimensions 'cues' and
+        'outcome_vector_dimensions'
     n_jobs : int
         an integer giving the number of threads in which the job should be
         executed
@@ -337,15 +370,16 @@ def _wh_binary_to_real(events, eta, outcome_vectors, *,
     temporary_directory : str
         path to directory to use for storing temporary files created;
         if none is provided, the operating system's default will
-        be used (/tmp on unix)
+        be used like '/tmp' on unix
     events_per_temporary_file: int
         Number of events in each temporary binary file. Has to be larger than 1
 
     Returns
     -------
     weights : xarray.DataArray
-        with dimensions 'vector dimensions' and 'cues'. You can lookup the weights
-        between a vector dimension and a cue with ``weights.loc[{'outcome_vector_dimensions': outcome_vector_dimension,
+        with dimensions 'outcome_vector dimensions' and 'cues'. You can lookup
+        the weights between a vector dimension and a cue with
+        ``weights.loc[{'outcome_vector_dimensions': outcome_vector_dimension,
         'cues': cue}]`` or ``weights.loc[outcome_vector_dimension].loc[cue]``.
 
     """
@@ -365,6 +399,8 @@ def _wh_binary_to_real(events, eta, outcome_vectors, *,
 
     if not outcome_vectors.data.data.c_contiguous:
         raise ValueError('outcome_vectors have to be c_contiguous')
+    if not outcome_vectors.dtype == np.float64:
+        raise ValueError('outcome_vectors have to be of dtype np.float64')
 
     # preprocessing
     n_events, cues, outcomes_from_events = count.cues_outcomes(events,
@@ -436,7 +472,8 @@ def _wh_binary_to_real(events, eta, outcome_vectors, *,
                                                     weights,
                                                     n_outcomes_per_job,
                                                     n_jobs)
-        # elif method == 'threading':
+        elif method == 'threading':
+            raise ValueError('TODO: for now: method needs "openmp"')
         #    part_lists = ndl.slice_list(all_outcome_indices, n_outcomes_per_job)
 
         #    working_queue = Queue(len(part_lists))
@@ -464,7 +501,7 @@ def _wh_binary_to_real(events, eta, outcome_vectors, *,
         #    for thread in threads:
         #        thread.join()
         else:
-            raise ValueError('method needs to be either "threading" or "openmp"')
+            raise ValueError('method needs to be either "openmp"')
 
     cpu_time_stop = time.process_time()
     wall_time_stop = time.perf_counter()
@@ -493,10 +530,10 @@ def _wh_real_to_binary(events, betas, lambda_, cue_vectors, *,
                        events_per_temporary_file=10000000):
     """
     Calculate the weights for all events using the Widrow-Hoff learning rule
-    and training as cue_vectors on outcomes.
+    from continuous cue vectors onto binary outcomes.
 
-    This is a parallel python implementation using numpy, multithreading and
-    the binary format defined in preprocess.py.
+    This is a parallel python implementation using cython and the binary format
+    defined in preprocess.py.
 
     Parameters
     ----------
@@ -509,7 +546,8 @@ def _wh_real_to_binary(events, betas, lambda_, cue_vectors, *,
         matrix that contains the cue vectors for each cue
     method : {'openmp', 'threading'}
     weights : None or xarray.DataArray
-        the xarray.DataArray needs to have the dimensions 'cues' and 'outcomes'
+        the xarray.DataArray needs to have the dimensions
+        'cue_vector_dimensions' and 'outcomes'
     n_jobs : int
         an integer giving the number of threads in which the job should be
         executed
@@ -525,7 +563,7 @@ def _wh_real_to_binary(events, betas, lambda_, cue_vectors, *,
     temporary_directory : str
         path to directory to use for storing temporary files created;
         if none is provided, the operating system's default will
-        be used (/tmp on unix)
+        be used like '/tmp' on unix
     events_per_temporary_file: int
         Number of events in each temporary binary file. Has to be larger than 1
 
@@ -551,6 +589,8 @@ def _wh_real_to_binary(events, betas, lambda_, cue_vectors, *,
 
     if not cue_vectors.data.data.c_contiguous:
         raise ValueError('cue_vectors have to be c_contiguous')
+    if not cue_vectors.dtype == np.float64:
+        raise ValueError('cue_vectors have to be of dtype np.float64')
 
     weights_ini = weights
     wall_time_start = time.perf_counter()
@@ -637,6 +677,9 @@ def _wh_real_to_binary(events, betas, lambda_, cue_vectors, *,
                                                     weights.data,
                                                     n_outcomes_per_job,
                                                     n_jobs)
+        else:
+            # TODO: implement threading
+            raise ValueError('TODO: for now: method needs to be "openmp"')
 
         weights = weights.reset_coords(drop=True)
 
@@ -665,10 +708,10 @@ def _wh_real_to_real(events, eta, cue_vectors, outcome_vectors, *,
                      events_per_temporary_file=10000000):
     """
     Calculate the weights for all events using the Widrow-Hoff learning rule
-    and training as outcomes on sematic vectors in semantics.
+    from cue vectors onto outcome vectors.
 
-    This is a parallel python implementation using numpy, multithreading and
-    the binary format defined in preprocess.py.
+    This is a parallel python implementation using cython, numpy,
+    multithreading and the binary format defined in preprocess.py.
 
     Parameters
     ----------
@@ -682,7 +725,8 @@ def _wh_real_to_real(events, eta, cue_vectors, outcome_vectors, *,
         matrix that contains the target vectors for each outcome
     method : {'openmp', 'threading', 'numpy'}
     weights : None or xarray.DataArray
-        the xarray.DataArray needs to have the dimensions 'cues' and 'outcomes'
+        the xarray.DataArray needs to have the dimensions
+        'cue_vector_dimensions' and 'outcome_vector_dimensions'
     n_jobs : int
         an integer giving the number of threads in which the job should be
         executed
@@ -698,7 +742,7 @@ def _wh_real_to_real(events, eta, cue_vectors, outcome_vectors, *,
     temporary_directory : str
         path to directory to use for storing temporary files created;
         if none is provided, the operating system's default will
-        be used (/tmp on unix)
+        be used like '/tmp' on unix
     events_per_temporary_file: int
         Number of events in each temporary binary file. Has to be larger than 1
 
@@ -733,9 +777,13 @@ def _wh_real_to_real(events, eta, cue_vectors, outcome_vectors, *,
 
     if not cue_vectors.data.data.c_contiguous:
         raise ValueError('cue_vectors have to be c_contiguous')
+    if not cue_vectors.dtype == np.float64:
+        raise ValueError('cue_vectors have to be of dtype np.float64')
 
     if not outcome_vectors.data.data.c_contiguous:
         raise ValueError('outcome_vectors have to be c_contiguous')
+    if not outcome_vectors.dtype == np.float64:
+        raise ValueError('outcome_vectors have to be of dtype np.float64')
 
     # preprocessing
     n_events, cues_from_events, outcomes_from_events = count.cues_outcomes(events,
